# -*- coding: utf8 -*-

import fcntl
import hashlib
import logging
import os
import signal
import struct
import sys
import traceback

try:
    from time import process_time as get_cpu_time  # Py3.3+
except ImportError:
    from time import clock as get_cpu_time

import posix_ipc

__all__ = ['ConsoleHandler', 'FileHandler']


class MultiprocessHandler(object):
    """
    Mixin pro logovací handler, který přidává funkčnost zamykání v
    multiprocess prostředí.
    """

    terminator = '\n'

    def emit(self, record):
        try:
            msg = self.format(record)
            stream = self.stream
            fcntl.lockf(self.stream, fcntl.LOCK_EX)
            try:
                stream.write(msg)
                stream.write(self.terminator)
                self.flush()
            finally:
                fcntl.lockf(self.stream, fcntl.LOCK_UN)
        except Exception:
            self.handleError(record)


class ConsoleHandler(MultiprocessHandler, logging.StreamHandler):
    """
    Handler posílající log zprávu do *sys.stdout* nebo *sys.stderr*. Pokud
    není specifikováno, výchozí stream je *sys.stderr*.

    ::

        handler_stdout = ConsoleHandler('stdout')
        handler_stderr = ConsoleHandler('stderr')
    """

    STREAM_MAP = {
        'stdout': sys.stdout,
        'stderr': sys.stderr,
    }

    def __init__(self, stream='stderr'):
        try:
            stream = self.STREAM_MAP[stream]
        except KeyError:
            raise ValueError("Unknown stream name `{}`".format(stream))
        super(ConsoleHandler, self).__init__(stream)


class FileHandler(MultiprocessHandler, logging.FileHandler):
    """
    Handler posílající log zprávu do souboru na disku. Argumenty konstruktoru
    mají stejný význam jako u třídy :class:`logging.FileHandler`.

    Handler reaguje na signál **SIGHUP**, při kterém nastaví příznak k
    uzavření souboru. K samotnému uzavření a znovuotevření dojde před
    zalogováním další zprávy.

    Argumenty mají stejný význam jako ve třídě ``logging.FileHandler``,
    *shmem_name* je identifikátor sdílené paměti, přes kterou handlery v
    multiprocesovém prostředí komunikují, musí to být „jednoúrovňová cesta“,
    např ``/my-app``. Pokud je :const:`None`, identifikátor bude odvozen z
    *filename*.
    """

    def __init__(self, filename, mode='a', encoding=None, delay=False,
                 shmem_name=None):
        super(FileHandler, self).__init__(filename, mode, encoding, delay)

        if shmem_name is None:
            sha1 = hashlib.sha1(filename.encode('utf-8'))
            shmem_name = '/{}'.format(sha1.hexdigest())
        self._rotate_flag_shm = posix_ipc.SharedMemory(
            shmem_name, flags=posix_ipc.O_CREAT, size=8)
        self._set_rotate_flag(0)
        self._last_cpu_time = 0

        self._previous_sighup_handler = signal.signal(
            signal.SIGHUP, self._sighup_handler)
        signal.siginterrupt(signal.SIGHUP, False)

    def emit(self, record):
        # Pokud se změní hodnota v _rotate_flag_shm, je to signál k
        # uzavření souboru.
        cpu_time = self._get_rotate_flag()
        if cpu_time > self._last_cpu_time:
            self._last_cpu_time = cpu_time
            self.close()

        if self.stream is None:
            self.stream = self._open()
        super(FileHandler, self).emit(record)

    def _sighup_handler(self, signum, frame):
        try:
            # Nastavim priznak k prerotovani
            self._set_rotate_flag(get_cpu_time())
            # Pokud existuje, zavolam dalsi signalhandler v rade
            if self._previous_sighup_handler:
                self._previous_sighup_handler(signum, frame)
        except Exception:
            sys.stderr.write(traceback.format_exc())
            sys.stderr.flush()

    def _get_rotate_flag(self):
        fcntl.lockf(self._rotate_flag_shm.fd, fcntl.LOCK_EX)
        try:
            os.lseek(self._rotate_flag_shm.fd, 0, os.SEEK_SET)
            data = os.read(self._rotate_flag_shm.fd, 8)
        finally:
            fcntl.lockf(self._rotate_flag_shm.fd, fcntl.LOCK_UN)
        if len(data) != 8:
            raise IOError('Shm read error')
        return struct.unpack('d', data)[0]

    def _set_rotate_flag(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("'value' must be a float type")
        data = struct.pack('d', value)
        fcntl.lockf(self._rotate_flag_shm.fd, fcntl.LOCK_EX)
        try:
            os.lseek(self._rotate_flag_shm.fd, 0, os.SEEK_SET)
            if os.write(self._rotate_flag_shm.fd, data) != 8:
                raise IOError('Shm write error')
        finally:
            fcntl.lockf(self._rotate_flag_shm.fd, fcntl.LOCK_UN)
